// Based on the work by DFRobot

#include "LiquidCrystal_I2C.h"
#include <inttypes.h>

// When the display powers up, it is configured as follows:
//
// 1. Display clear
// 2. Function set:
//    DL = 1; 8-bit interface data
//    N = 0; 1-line display
//    F = 0; 5x8 dot character font
// 3. Display on/off control:
//    D = 0; Display off
//    C = 0; Cursor off
//    B = 0; Blinking off
// 4. Entry mode set:
//    I/D = 1; Increment by 1
//    S = 0; No shift
//
// Note, however, that resetting the Arduino doesn't reset the LCD, so we
// can't assume that its in that state when a sketch starts (and the
// LiquidCrystal constructor is called).

LiquidCrystal_I2C::LiquidCrystal_I2C(uint8_t lcd_Addr,uint8_t lcd_cols,uint8_t lcd_rows)
{
  _Addr = lcd_Addr;
  _cols = lcd_cols;
  _rows = lcd_rows;
  _backlightval = LCD_NOBACKLIGHT;
}

void LiquidCrystal_I2C::oled_init(){
  _oled = true;
    init_priv();
}

void LiquidCrystal_I2C::init(){
    init_priv();
}

void LiquidCrystal_I2C::init_priv()
{
    //enable I2C module 0
    SysCtlPeripheralEnable(SYSCTL_PERIPH_I2C0);

    //reset module
    SysCtlPeripheralReset(SYSCTL_PERIPH_I2C0);

    //enable GPIO peripheral that contains I2C 0
    SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOB);

    // Configure the pin muxing for I2C0 functions on port B2 and B3.
    GPIOPinConfigure(GPIO_PB2_I2C0SCL);
    GPIOPinConfigure(GPIO_PB3_I2C0SDA);

    // Select the I2C function for these pins.
    GPIOPinTypeI2CSCL(GPIO_PORTB_BASE, GPIO_PIN_2);
    GPIOPinTypeI2C(GPIO_PORTB_BASE, GPIO_PIN_3);

    // Enable and initialize the I2C0 master module.  Use the system clock for
    // the I2C0 module.  The last parameter sets the I2C data transfer rate.
    // If false the data rate is set to 100kbps and if true the data rate will
    // be set to 400kbps.
    I2CMasterInitExpClk(I2C0_BASE, SysCtlClockGet(), false);

    //clear I2C FIFOs
    HWREG(I2C0_BASE + I2C_O_FIFOCTL) = 80008000;

    _displayfunction = LCD_4BITMODE | LCD_1LINE | LCD_5x8DOTS;
    begin(_cols, _rows);
}

void LiquidCrystal_I2C::begin(uint8_t cols, uint8_t lines, uint8_t dotsize) {
    if (lines > 1) {
        _displayfunction |= LCD_2LINE;
    }
    _numlines = lines;

    // for some 1 line displays you can select a 10 pixel high font
    if ((dotsize != 0) && (lines == 1)) {
        _displayfunction |= LCD_5x10DOTS;
    }

    this->delay(50);
    // Now we pull both RS and R/W low to begin commands
    expanderWrite(_backlightval);   // reset expanderand turn backlight off (Bit 8 =1)
    this->delay(1000);

    //put the LCD into 4 bit mode
    // this is according to the hitachi HD44780 datasheet
    // figure 24, pg 46

      // we start in 8bit mode, try to set 4 bit mode
   write4bits(0x03 << 4);
   this->delayMicroseconds(4500);

   // second try
   write4bits(0x03 << 4);
   this->delayMicroseconds(4500);

   // third go!
   write4bits(0x03 << 4);
   this->delayMicroseconds(4500);

   // finally, set to 4-bit interface
   write4bits(0x02 << 4);


    // set # lines, font size, etc.
    command(LCD_FUNCTIONSET | _displayfunction);

    // turn the display on with no cursor or blinking default
    _displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;
    display();

    // clear it off
    clear();

    // Initialize to default text direction (for roman languages)
    _displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;

    // set the entry mode
    command(LCD_ENTRYMODESET | _displaymode);

    home();

}

/********** high level commands, for the user! */
void LiquidCrystal_I2C::clear(){
    command(LCD_CLEARDISPLAY);// clear display, set cursor position to zero
    delayMicroseconds(2000);
    if (_oled) setCursor(0,0);
}

void LiquidCrystal_I2C::home(){
    command(LCD_RETURNHOME);
    delayMicroseconds(2000); // set cursor position to zero
}

void LiquidCrystal_I2C::setCursor(uint8_t col, uint8_t row){
    int row_offsets[] = { 0x00, 0x40, 0x14, 0x54 };
    if ( row > _numlines ) {
        row = _numlines-1;    // we count rows starting w/0
    }
    command(LCD_SETDDRAMADDR | (col + row_offsets[row]));
}

// Turn the display on/off (quickly)
void LiquidCrystal_I2C::noDisplay() {
    _displaycontrol &= ~LCD_DISPLAYON;
    command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void LiquidCrystal_I2C::display() {
    _displaycontrol |= LCD_DISPLAYON;
    command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// Turns the underline cursor on/off
void LiquidCrystal_I2C::noCursor() {
    _displaycontrol &= ~LCD_CURSORON;
    command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void LiquidCrystal_I2C::cursor() {
    _displaycontrol |= LCD_CURSORON;
    command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// Turn on and off the blinking cursor
void LiquidCrystal_I2C::noBlink() {
    _displaycontrol &= ~LCD_BLINKON;
    command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void LiquidCrystal_I2C::blink() {
    _displaycontrol |= LCD_BLINKON;
    command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// These commands scroll the display without changing the RAM
void LiquidCrystal_I2C::scrollDisplayLeft(void) {
    command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVELEFT);
}
void LiquidCrystal_I2C::scrollDisplayRight(void) {
    command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVERIGHT);
}

// This is for text that flows Left to Right
void LiquidCrystal_I2C::leftToRight(void) {
    _displaymode |= LCD_ENTRYLEFT;
    command(LCD_ENTRYMODESET | _displaymode);
}

// This is for text that flows Right to Left
void LiquidCrystal_I2C::rightToLeft(void) {
    _displaymode &= ~LCD_ENTRYLEFT;
    command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'right justify' text from the cursor
void LiquidCrystal_I2C::autoscroll(void) {
    _displaymode |= LCD_ENTRYSHIFTINCREMENT;
    command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'left justify' text from the cursor
void LiquidCrystal_I2C::noAutoscroll(void) {
    _displaymode &= ~LCD_ENTRYSHIFTINCREMENT;
    command(LCD_ENTRYMODESET | _displaymode);
}

// Turn the (optional) backlight off/on
void LiquidCrystal_I2C::noBacklight(void) {
    _backlightval=LCD_NOBACKLIGHT;
    expanderWrite(0);
}

void LiquidCrystal_I2C::backlight(void) {
    _backlightval=LCD_BACKLIGHT;
    expanderWrite(0);
}

/*********** mid level commands, for sending data/cmds */

inline void LiquidCrystal_I2C::command(uint8_t value) {
    send(value, 0);
}


/************ low level data pushing commands **********/

// write either command or data
void LiquidCrystal_I2C::send(uint8_t value, uint8_t mode) {
    uint8_t highnib=value&0xf0;
    uint8_t lownib=(value<<4)&0xf0;
    write4bits((highnib)|mode);
    write4bits((lownib)|mode);
}

void LiquidCrystal_I2C::write4bits(uint8_t value) {
    expanderWrite(value);
    delayMicroseconds(1);

    pulseEnable(value);
    delayMicroseconds(50);
}

void LiquidCrystal_I2C::expanderWrite(uint8_t _data){
    // Tell the master module what address it will place on the bus when
    // communicating with the slave.
    I2CMasterSlaveAddrSet(I2C0_BASE, _Addr, false);

    //put data to be sent into FIFO
    I2CMasterDataPut(I2C0_BASE, (int)(_data) | _backlightval);

    //Initiate send of data from the MCU
    I2CMasterControl(I2C0_BASE, I2C_MASTER_CMD_SINGLE_SEND);

    // Wait until MCU is done transferring.
    while(I2CMasterBusy(I2C0_BASE));
}

void LiquidCrystal_I2C::pulseEnable(uint8_t _data){
    expanderWrite(_data | En);  // En high

    expanderWrite(_data & ~En); // En low
}

void LiquidCrystal_I2C::cursor_on(){
    cursor();
}

void LiquidCrystal_I2C::cursor_off(){
    noCursor();
}

void LiquidCrystal_I2C::blink_on(){
    blink();
}

void LiquidCrystal_I2C::blink_off(){
    noBlink();
}

void LiquidCrystal_I2C::setBacklight(uint8_t new_val){
    if(new_val){
        backlight();        // turn backlight on
    }else{
        noBacklight();      // turn backlight off
    }
}

void LiquidCrystal_I2C::delay(uint32_t ui32Ms)
{
    SysCtlDelay(ui32Ms * (SysCtlClockGet() / 3 / 1000));
}

void LiquidCrystal_I2C::delayMicroseconds(uint32_t ui32Us)
{
    SysCtlDelay(ui32Us * (SysCtlClockGet() / 3 / 1000000));
}

void LiquidCrystal_I2C::print(char array[])
{
    for(int i=0; array[i] != '\0'; i++)
    {
        this->send(array[i],Rw);
    }

//    // Tell the master module what address it will place on the bus when
//    // communicating with the slave.
//    I2CMasterSlaveAddrSet(I2C0_BASE, _Addr, false);
//
//    //put data to be sent into FIFO
//    I2CMasterDataPut(I2C0_BASE, array[0]);
//
//    //if there is only one argument, we only need to use the
//    //single send I2C function
//    if(array[1] == '\0')
//    {
//        //Initiate send of data from the MCU
//        I2CMasterControl(I2C0_BASE, I2C_MASTER_CMD_SINGLE_SEND);
//
//        // Wait until MCU is done transferring.
//        while(I2CMasterBusy(I2C0_BASE));
//    }
//
//    //otherwise, we start transmission of multiple bytes on the
//    //I2C bus
//    else
//    {
//        //Initiate send of data from the MCU
//        I2CMasterControl(I2C0_BASE, I2C_MASTER_CMD_BURST_SEND_START);
//
//        // Wait until MCU is done transferring.
//        while(I2CMasterBusy(I2C0_BASE));
//
//        //initialize index into array
//        uint8_t i = 1;
//
//        //send num_of_args-2 pieces of data, using the
//        //BURST_SEND_CONT command of the I2C module
//        while(array[i + 1] != '\0')
//        {
//            //put next piece of data into I2C FIFO
//            I2CMasterDataPut(I2C0_BASE, array[i++]);
//
//            //send next data that was just placed into FIFO
//            I2CMasterControl(I2C0_BASE, I2C_MASTER_CMD_BURST_SEND_CONT);
//
//            // Wait until MCU is done transferring.
//            while(I2CMasterBusy(I2C0_BASE));
//        }
//
//        //put last piece of data into I2C FIFO
//        I2CMasterDataPut(I2C0_BASE, array[i]);
//
//        //send next data that was just placed into FIFO
//        I2CMasterControl(I2C0_BASE, I2C_MASTER_CMD_BURST_SEND_FINISH);
//
//        // Wait until MCU is done transferring.
//        while(I2CMasterBusy(I2C0_BASE));
//    }
}
